@@ -22,9 +22,11 @@ class _Node:
     types: List[_NodeType]
     optional: bool = True
     is_list: bool = False
+    nested_lists: int = 1
     parent: Optional["_Node"] = None
     children: List["_Node"] = field(default_factory=list)
     class_name: Optional[str] = None
+    final_class_name: Optional[str] = None
     extends: Optional["_Node"] = None
 
     def path(self):
@@ -43,13 +45,15 @@ def _determine_if_optional(values: Set) -> bool:
     return False
 
 
-def _generate_class_for_node(current_node, class_map):
+def _generate_class_for_node(current_node: _Node, class_map):
     if _NodeType.CLASS in current_node.types:
         class_name = current_node.class_name
+        if current_node.final_class_name is not None:
+            class_name = current_node.final_class_name
         fields = []
         for child in current_node.children:
             fields.append(child.name + ": " + _get_node_class_type(child))
-        fields_string = "\n\t".join(fields)
+        fields_string = "\n\t".join(sorted(fields))
         extends = ""
         if current_node.extends is not None:
             extends = ", " + current_node.extends.class_name
@@ -58,12 +62,12 @@ class {class_name}(Protocol{extends}):
 \t{fields_string}
 
 """
-        class_map[current_node.class_name] = result
+        class_map[class_name] = result
 
 
 def _get_node_class_type(child: _Node):
     def base_value(types):
-        if child.optional and len(types) == 1:
+        if child.optional and ((len(types) == 1 and types[0] != "None") or (len(types) == 2 and "None" in types)):
             return f"Optional[{types[0]}]"
         if len(types) > 1:
             return "Union[" + (",".join(types)) + "]"
@@ -71,7 +75,10 @@ def _get_node_class_type(child: _Node):
 
     result = []
     if _NodeType.CLASS in child.types:
-        result.append(f'"{child.class_name}"')
+        name = child.class_name
+        if child.final_class_name is not None:
+            name = child.final_class_name
+        result.append(f'"{name}"')
     if _NodeType.NUMBER in child.types:
         result.append("int")
     if _NodeType.STRING in child.types:
@@ -82,11 +89,26 @@ def _get_node_class_type(child: _Node):
         result.append("None")
 
     if child.is_list:
-        return f"List[{base_value(result)}]"
+        list_string = ""
+        for each in range(child.nested_lists):
+            list_string += "List["
+        list_string += base_value(result)
+        for each in range(child.nested_lists):
+            list_string += "]"
+        return list_string
     else:
         return base_value(result)
 
 
+def _get_name_to_node_map(node_map):
+    result = {}
+    for _, node in node_map.items():
+        if result.get(node.class_name) is None:
+            result[node.class_name] = []
+        result[node.class_name].append(node)
+    return result
+
+
 def _generate_classes(node_map: Dict[str, _Node], root_name: str):
     # find root
     # set node class names
@@ -105,24 +127,32 @@ def _generate_classes(node_map: Dict[str, _Node], root_name: str):
     root.name = root_name[0].lower() + root_name[1:]
     # rename duplicates with different properties
     # extends duplicates when its properties are subsets of another of same name
+    name_to_node_map = _get_name_to_node_map(node_map)
+    class_usage_count: Dict[str, int] = {}
     for _, node in node_map.items():
-        for _, other_node in node_map.items():
-            if node != other_node and node.class_name == other_node.class_name \
-                    and _NodeType.CLASS in node.types and _NodeType.CLASS in other_node.types:
-                children_names = set(map(lambda x: x.class_name, node.children))
-                other_children_names = set(map(lambda x: x.class_name, other_node.children))
+        for other_node in name_to_node_map[node.class_name]:  # type:_Node
+            if node != other_node and _NodeType.CLASS in node.types and _NodeType.CLASS in other_node.types:
+                children_names = set([x.name for x in node.children])
+                other_children_names = set([x.name for x in other_node.children])
                 if len(children_names) != len(other_children_names):
                     if set.issubset(children_names, other_children_names):
                         node.extends = other_node
                     elif set.issubset(other_children_names, children_names):
                         other_node.extends = node
-                if node.extends is None and other_node.extends is None and len(children_names) != len(
-                        other_children_names):
+                if node.extends is None and other_node.extends is None and len(
+                        children_names.intersection(other_children_names)) == 0 and other_node.final_class_name is None:
                     assert other_node.class_name is not None
-                    other_node.class_name = other_node.class_name + "2"
+                    if class_usage_count.get(other_node.class_name) is None:
+                        class_usage_count[other_node.class_name] = 0
+                    class_usage_count[other_node.class_name] = class_usage_count[other_node.class_name] + 1
+                    if class_usage_count[other_node.class_name] == 1:
+                        other_node.final_class_name = other_node.class_name
+                    else:
+                        other_node.final_class_name = other_node.class_name + str(
+                            class_usage_count[other_node.class_name])
 
     # Depth first traversal while creating classes
-    class_map:Dict[str, str] = {}
+    class_map: Dict[str, str] = {}
 
     def loop(current_node: _Node):
         _generate_class_for_node(current_node, class_map)
@@ -167,7 +197,7 @@ def _generate_classes_text(file_name, parsed_json):
     class_name = file_name
     if "_" in class_name:
         class_name = class_name.replace("_", " ").title().replace(" ", "")
-    json_map = _dict_to_dot_map(parsed_json, class_name)
+    json_map = _dict_to_dot_map({class_name: parsed_json}, class_name)
     #   iterate over key and objects keeping track of parentage
     parent_dict = _create_parent_dictionary(json_map)
     # build node tree
@@ -187,10 +217,10 @@ def _normalize(map_or_list_of_maps: Union[List, Dict[str, Any]]):
         # get immutalbe list
         keys = tuple(map_or_list_of_maps.keys())
         for key in keys:
-            if " " in key:
+            if re.search(pattern=r"[ \.]", string=key):
                 value = map_or_list_of_maps[key]
                 map_or_list_of_maps.pop(key, None)
-                map_or_list_of_maps[key.replace(" ", "_")] = value
+                map_or_list_of_maps[re.sub(pattern=r"[ \.]", string=key, repl="_")] = value
         for key, value in map_or_list_of_maps.items():
             _normalize(value)
 
@@ -226,15 +256,17 @@ def _create_parent_dictionary(map):
 
     # depth first search to aggregate all possible values and paths
     def loop(obj, parent_string: str):
-        if isinstance(obj, DotMap):
+        if isinstance(obj, DotMap) or isinstance(obj, dict):
             properties = dir(obj)
+            if isinstance(obj, dict):
+                properties = obj.keys()
             for prop in properties:
                 child = obj[prop]
                 if parent_string.strip() == "":
                     new_parent = prop
                 else:
                     new_parent = parent_string + "." + prop
-                if isinstance(child, DotMap) or isinstance(child, list):
+                if isinstance(child, DotMap) or isinstance(child, list) or isinstance(child, dict):
                     loop(child, new_parent)
                 else:
                     if parent_dict.get(new_parent) is None:
@@ -244,7 +276,7 @@ def _create_parent_dictionary(map):
         elif isinstance(obj, list):
             for index, child in enumerate(obj):
                 list_parent_string = parent_string + "[x]"
-                if isinstance(child, DotMap) or isinstance(child, list):
+                if isinstance(child, DotMap) or isinstance(child, list) or isinstance(child, dict):
                     loop(child, list_parent_string)
                 else:
                     if parent_dict.get(list_parent_string) is None:
@@ -252,6 +284,14 @@ def _create_parent_dictionary(map):
                     parent_dict[list_parent_string].add(type(child))
 
     loop(map, "")
+    # cleanup situations where a list could also be null
+    to_test = [key for key, x in parent_dict.items() if type(None) in x]
+    # remove these keys if there is a list version of them
+    parent_keys = [k for k, y in parent_dict.items()]
+    to_remove = [key for key in to_test if len([y for y in parent_keys if f"{key}[x]" in y]) > 0]
+    [parent_dict.pop(x) for x in to_remove]
+    # mark the optional lists are nullable
+    # TODO
     return parent_dict
 
 
@@ -263,20 +303,23 @@ def _create_node_map(parent_dict):
             node_key = ".".join(split[0:index + 1])
             # if not exists add new
             parent_node = None
-            node_types, is_list = _determine_type(index, split, value)
+            node_types, is_list, nested_lists = _determine_type(index, split, value)
             name = split[index].replace("[x]", "")
             optional = _determine_if_optional(value)
             if index > 0:
                 parent_node = node_map[".".join(split[0:index])]
             if node_map.get(node_key) is None:
-                node_map[node_key] = _Node(name=name, types=node_types, optional=optional, is_list=is_list)
+                node_map[node_key] = _Node(name=name, types=node_types, optional=optional, is_list=is_list,
+                                           nested_lists=nested_lists)
+            elif optional and index==len(split)-1:
+                node_map[node_key].optional = optional
             if parent_node and node_map[node_key].parent is None:
                 node_map[node_key].parent = parent_node
                 parent_node.children.append(node_map[node_key])
     return node_map
 
 
-def _determine_type(index, split, value: Set) -> Tuple[List[_NodeType], bool]:
+def _determine_type(index, split, value: Set) -> Tuple[List[_NodeType], bool, int]:
     node_types = []
     is_list = False
     if "[x]" in split[index]:
@@ -304,21 +347,16 @@ def _determine_type(index, split, value: Set) -> Tuple[List[_NodeType], bool]:
                 node_types.append(_NodeType.BOOLEAN)
             if _has_type(value, None):
                 node_types.append(_NodeType.NULL)
-
-    return node_types, is_list
-
-
-def _all_of_type(array: Union[List, Set], clazz):
-    for value in array:
-        if value is not type(None) and value is not clazz:
-            return False
-    return True
+    nested_lists = len(re.findall(pattern=r"\[x\]", string=split[index]))
+    return node_types, is_list, nested_lists
 
 
 def _has_type(array: Union[List, Set], clazz):
     for value in array:
         if value is not type(None) and value is clazz:
             return True
+        if value is type(None) and value is type(clazz):
+            return True
     return False
 
 
